/*********************************************************************/
/* data.c                                                            */
/* function neblo_handle_RES_msgs()                                 */
/* see COPYRIGHT in parent directory                                 */
/*********************************************************************/

#include <neblo_.h>
#include <neblo_macros.h>
#include <neblo_secret.h>
#include <queues.h>


/*
 * this peer is asked to take a RESponse msg, and deliver to application
 * if this peer is the final destination, or propagate towards source otherwise
 */
int
neblo_handle_RES_msgs
( reqst_t *reqst )
{

  unsigned int          rport;
  neblo_port            port;
  unsigned int          len;
  neblo_rhandle         handle;
  neblo_addr            addr;
  neblo_addr            real_addr;
  pend_read_t          *pr;
  int                   rc;


  sscanf(reqst->buf,"%*s %04x %08x %08x %016llx %016llx",
         &rport,&len,&handle,&addr,&real_addr);
  port = (neblo_port)rport;

  /*
   * SECURITY: requests of this kind may only be accepted
   * if coming back from a neighbour or finger, or when they arise as
   * answers to locally-originated retried read msg (no sock)
   */
  if (reqst->csock != NULL
      && !is_neighbour_or_finger((peer_t *)is_friend(reqst->csock))) {

#ifdef DEBUG_RW
    S(fprintf(stderr,"[%lu]\tREFUSED data: not neigh/finger/local\n",
              reqst->th_id));
#endif

    return -1;
  }

  pr = search_read_reqst(handle);

  if (pr == NULL) {

#ifdef DEBUG_RW
    S(fprintf(stderr,"[%lu]\tREFUSED data: handle not found\n",reqst->th_id));
#endif

    return -1;
  }

  pthread_mutex_lock(&pr->lock);

  if (pr->addr != real_addr) {
    /* pending read does not match. discard and quit service */

#ifdef DEBUG_RW
    S(fprintf(stderr,"[%lu]\tREFUSED data: no match\n",reqst->th_id));
#endif

    /*  pending read "pr" remains in the queue.  */

    pthread_mutex_unlock(&pr->lock);
    return -1;
  }

  if (pr->done > 0) {
    /* read msg already successfully served, so discard and quit */

#ifdef DEBUG_RW
    S(fprintf(stderr,
            "[%lu]\tread msg %u already served (pr %u)\n",
            reqst->th_id,(unsigned int)handle,(unsigned int)pr));
#endif

    pthread_mutex_unlock(&pr->lock);
    return 0;
  }

  /* trunc to max allowed size for the matching read msg */
  len = min(len,pr->maxlen);

  if (pr->from == NULL) {
    /* read msg originated from this peer */
    if (len) {
      /* pass data up to application, then quit */
      memcpy(pr->data,reqst->buf+HEADERSIZE_R,len);
      pr->done = 1;
      pthread_cond_signal(&pr->wait_cond);
      /* neblo_RES_wait()/neblo_RES_test() will cancel the request from queue */
    } else {
      /* zero-sized data means "data not found" */
      pr->done = -1;
    }
    pthread_mutex_unlock(&pr->lock);
    return 0;
  }

  /* read originated from a remote peer.  backtrack the route */

  invoke_forwarder_handler_RES(port,real_addr,reqst->buf+HEADERSIZE_R,len);

  rc = forward_msg(DATA,pr->from,port,addr,real_addr,reqst->buf,len,
                   pr->from_rhandle,0,reqst->th_id);

  if (rc >= 0) {
    /* successfully forwarded back, so cancel handle */
    cancel_read_reqst(pr);
    pthread_mutex_unlock(&pr->lock);
    return rc;
  }

#ifdef DEBUG_RW
  S(fprintf(stderr,
            "[%lu]\tforward_msg() to addr %016llx (real addr %016llx) returned %d\n",
            reqst->th_id,addr,real_addr,rc));
#endif

  pthread_mutex_unlock(&pr->lock);
  return rc;

}
