/*********************************************************************/
/* findsucc.c                                                        */
/* function neblo_findsucc()                                         */
/* see COPYRIGHT in parent directory                                 */
/*********************************************************************/

#include <neblo_.h>
#include <neblo_macros.h>
#include <neblo_secret.h>
#include <queues.h>


/*
 *  This peer is asked to become the new successor of requestor.
 *  Command:
 *  findsucc <handle> <orig_ip_addr> <orig_ip_port> <orig_pubkey_mod>
 *           <orig_pubkey_exp>
 *           [subcommands: ok, end <my_id>, forwarded]
 */
int
neblo_findsucc
( reqst_t *reqst )
{

  unsigned long         raddr;
  struct in_addr        originator_addr;
  unsigned int          rport;
  u_int16_t             originator_port;
  u_int32_t             cmd_handle_val;
  char                 *originator_keymod;
  char                 *originator_keyexp;
  RSA                  *originator_pubkey = NULL;
  peer_t               *f,*ff;
  csocket_t            *cs = NULL;
  unsigned int          i;
  int                   rc = 0;


  _do_(_malloc_(originator_keymod,KEYBUFLEN+SLACK),
       !originator_keymod, "neblo_findsucc(): originator_keymod malloc");
  _do_(_malloc_(originator_keyexp,BUFLEN+SLACK),
       !originator_keyexp, "neblo_findsucc(): originator_keyexp malloc");

  sscanf(reqst->buf,"%*s %u %lu %u %50000s %100s",
         &cmd_handle_val,&raddr,&rport,originator_keymod,originator_keyexp);

  originator_addr.s_addr = raddr;
  originator_port = (u_int16_t)rport;


  /*  SECURITY:  accept req only if coming from a friend.  */

  if ((f = (peer_t *)is_friend(reqst->csock)) == NULL) {

#ifdef DEBUG_F
    S(fprintf(stderr,"[%lu]\tfindsucc req not from a friend.  abort...\n",
              reqst->th_id));
#endif

    goto abort;

  }

  if ((ff = predecessor()) == NULL) {

dead_predecessor:

    /*
     *  No predecessor or DEAD predecessor.  OK to become new successor of
     *  the requestor peer.  Establish a backward conn to it, and mark it
     *  as predecessor.  Distinguish between a direct req and a forwarded
     *  req (the latter may only arrive through the successor peer ).
     */
    if (!from_successor(reqst->csock)) {

      /*  Direct request, fast answer  */

      _write_(reqst->csock,"ok",strlen("ok")+1,rc,goto abort);

#ifdef DEBUG
      S(fprintf(stderr,"[%lu]\t<< ok\n",reqst->th_id));
#endif

      /*  Wait requestor to be ready to get our connection...  */
      _read_(reqst->csock,reqst->buf,BUFLEN,rc,TIMEOUT,goto abort);
      if (strncmp(reqst->buf,"ok",strlen("ok"))) {
        rc = -1;
        goto abort;
      }

#ifdef DEBUG
      S(fprintf(stderr,"[%lu]\t>> ok\n",reqst->th_id));
#endif

      /*  Connect to new predecessor  */
      rc = call_peer(f,2 /* from successor */,0,reqst->th_id);
      if (rc < 0) {

#ifdef DEBUG_M
        S(fprintf(stderr,
                  "[%lu]\tfailed to connect to new predecessor. abort...\n",
                  reqst->th_id));
#endif

        goto abort;

      }

      /*
       *  Tell predecessor we have connected to it, and give the
       *  lower bound of local addr interval, so new predecessor can
       *  update its own addr interval.
       */
      sprintf(reqst->buf,"end %016llx",ring->my_id);
      _write_(reqst->csock,reqst->buf,strlen(reqst->buf)+1,rc,goto abort);

#ifdef DEBUG
      S(fprintf(stderr,"[%lu]\t<< %s\n",reqst->th_id,reqst->buf));
#endif

      forall_friends(ff,
        if (ff->from_predecessor) {
          ff->from_predecessor = 0;
          break;
        }
      );
      f->from_predecessor = 1;

      /*  Done.  */

#ifdef DEBUG_M
      S(fprintf(stderr,"[%lu]\tconnected to new predecessor\n",reqst->th_id));
#endif

      goto exit;

    }

    /*
     *  Forwarded request, answer involves "newsucc" cmd.
     *  Open a temp connection to originator, challenge it for the pubkey,
     *  wait for it to become a friend, mark the friend as predecessor,
     *  then open a backward connection to the friend.
     *  On failures, do not close the connection from forwarder.
     */

    _do_(_RSA_new_(originator_pubkey),!originator_pubkey,
         "neblo_findsucc(): RSA_new of originator_pubkey");
    _convert_pubkey_(originator_pubkey,originator_keymod,originator_keyexp,
                     goto exit);
    _openout_(originator_addr,originator_port,&cs,originator_pubkey,&rc);
    if (rc < 0)
      goto exit;

    sprintf(reqst->buf,"newsucc %u %u %016llx %s %s",
            cmd_handle_val,PORT,ring->my_id,pubkeymod,pubkeyexp);
    _write_(cs,reqst->buf,strlen(reqst->buf)+1,rc,goto exit);

#ifdef DEBUG_M
    fprintf(stderr,"[%lu]\t<< %s\n",reqst->th_id,reqst->buf);
#endif

    /*  SECURITY II:  challenge originator about its supposed pubkey.  */

    rc = _put_challenge_(cs,reqst->buf,originator_pubkey);
    if (rc < 0) {

#ifdef DEBUG_M
      fprintf(stderr,"[%lu]\tchallenge interrupted\n",reqst->th_id);
#endif

      goto exit;
    }

#ifdef DEBUG_M
    fprintf(stderr,"[%lu]\tchallenge ok\n",reqst->th_id);
#endif

    /*  Tell to originator we are ready to get its connection */

    insert_expd_reqst(originator_addr,originator_port,1 /* from predecessor */,
                      originator_pubkey,reqst->th_id);
    _write_(cs,"ok",strlen("ok")+1,rc,goto abort);

#ifdef DEBUG
    S(fprintf(stderr,"[%lu]\t<< ok\n",reqst->th_id));
#endif

    /*
     *  The incoming connection will be managed by a separate thread.
     *  Predecessor tells us "ok" once it has established the conn with
     *  us and is ready to accept the connection in the opposite way.
     */
    _read_(cs,reqst->buf,BUFLEN,rc,TIMEOUT,goto exit);
    if (strncmp(reqst->buf,"ok",strlen("ok"))) {
      rc = -1;
      goto exit;
    }

#ifdef DEBUG
    S(fprintf(stderr,"[%lu]\t>> ok\n",reqst->th_id));
#endif

    /*  Connect to originator, that is, our new predecessor  */

    forall_friends(ff,if (ff->from_predecessor) {f=ff; break;});
    if (f == NULL) {  /* should never occur */
      rc = -1;
      goto exit;
    }

    for (i=5; i; i--) {
      rc = call_peer(f,2 /* from successor */,0,reqst->th_id);
      if (rc >= 0)
        /* success. predecessor->status is set to OK */
        break;
      if (rc == -2) {
        /* predecessor has died meanwhile. abort... */

#ifdef DEBUG_M
        S(fprintf(stderr,"\tpredecessor dead. abort...\n"));
#endif

        goto exit;
      }
      /* new neighbour not responding. wait a moment, then try again */
      sleep(1);
    }
    if (!i) {
      /* predecessor not responding for too long.  assume it is dead */

#ifdef DEBUG_M
        S(fprintf(stderr,"[%lu]\tpredecessor not responding, abort...\n",
                  reqst->th_id));
#endif

      rc = -1;
      goto exit;
    }

    /*  Done.  */

#ifdef DEBUG_M
    S(fprintf(stderr,
              "[%lu]\tneblo_findsucc(): new predecessor: %u.%u.%u.%u:%u\n",
              reqst->th_id,
              ((unsigned char *)&f->addr.s_addr)[0],
              ((unsigned char *)&f->addr.s_addr)[1],
              ((unsigned char *)&f->addr.s_addr)[2],
              ((unsigned char *)&f->addr.s_addr)[3],
              f->port));
#endif

    goto exit;

  }

  /*
   *  Predecessor not DEAD; maybe requestor simply does not know about it.
   *  In such case forward req backward to local predecessor, then tell to
   *  requestor to wait for another peer to respond (recursive management).
   */

  rc = tell_peer(ff,reqst->buf,strlen(reqst->buf)+1,0,reqst->th_id);
  if (rc < 0) {
    /*
     *  Uhm, predecessor has troubles.  Maybe it is dead actually.
     *  Close the ring here, as in the case of dead predecessor.
     */
    goto dead_predecessor;
  }

#ifdef DEBUG_M
  S(fprintf(stderr,"[%lu]\t<< %s\n",reqst->th_id,reqst->buf));
#endif

  _write_(reqst->csock,"forwarded",strlen("forwarded")+1,rc,goto abort);

#ifdef DEBUG_M
  S(fprintf(stderr,"[%lu]\t<< forwarded\n",reqst->th_id));
#endif


exit:

  _free_(originator_keymod);
  _free_(originator_keyexp);
  _RSA_free_(originator_pubkey);
  _close_(&cs);
  return rc;

abort:

  _close_(&reqst->csock);
  goto exit;

}
