/*********************************************************************/
/* neighlist.c                                                       */
/* function neblo_neighlist()                                        */
/* see COPYRIGHT in parent directory                                 */
/*********************************************************************/

#include <neblo_.h>
#include <neblo_macros.h>
#include <neblo_secret.h>
#include <queues.h>


/*
 * neblo_neighlist:  get the (possibly updated) list of neighbours
 * from the successor peer; this allows to update the local neighbour list
 * (command "neighlist").
 * Elements of the neighbour list are sent one after the other, in a
 * synchronous way; the cmd is thus managed in a bidirectional way.
 * Returns >=0 if success, < 0 otherwise.
 */
int 
neblo_neighlist
( reqst_t *reqst ) 
{
  
  unsigned long         faddr;
  unsigned int          fp;
  u_int16_t             fport;
  int                   rc;
  peer_t               *found;
  peer_t               *f;
  char                 *keymod;
  char                 *keyexp;
  peer_t               *insert_point;
  int                   i;


  /*  SECURITY:  accept cmd only if coming from our successor  */

  if (!from_successor(reqst->csock)) {

#ifdef DEBUG_R
    S(fprintf(stderr,"\tneblo_neighlist(): not from successor. abort...\n"));
#endif

    return -1;
  }

  _do_(_malloc_(keymod,KEYBUFLEN+SLACK),
       !keymod,"neblo_neighlist(): keymod malloc");
  _do_(_malloc_(keyexp,BUFLEN+SLACK),
       !keyexp,"neblo_neighlist(): keyexp malloc");

  /*  Tell successor we are ready to get the list  */
  _write_(reqst->csock,"ok",strlen("ok")+1,rc,goto exit);

#ifdef DEBUG_R
  S(fprintf(stderr, "\t<< ok\n"));
#endif

  /*
   *  Successor sends its current neighbour list.
   *  Based on this we try to update the local neighbour list, starting
   *  from 2nd element (i.e., next to successor; successor remains unchanged).
   */

  pthread_mutex_lock(&ring->lock);
  insert_point = ring->neigh_list;  /* insert next to successor */
  pthread_mutex_unlock(&ring->lock);
  /*
   *  The head of neigh list may be concurrently changed by a "join", but
   *  this would not affect the part of neighbour list we are now working at,
   *  so no need to assert the ring->lock during the entire list update.
   *  Also the neblo_check_neighbours() thread might want to act upon the
   *  neighbour list, but the "dont_clean" flag has been set when invoking
   *  this routine, so neighbours cannot be concurrently deleted from the
   *  list --> no race conditions
   */

  for (i=0;i<MAXNEIGH;i++) {

    _read_(reqst->csock,reqst->buf,KEYBUFLEN,rc,TIMEOUT,break);

    if (!strncmp(reqst->buf,"end",strlen("end"))) {

#ifdef DEBUG_R
      S(fprintf(stderr, "\t>> end\n"));
#endif

      /* end of list */
      rc = 0;
      break;
    }

    sscanf(reqst->buf, "%lu %u", &faddr,&fp); fport = (u_int16_t)fp;

#ifdef DEBUG_R
    S(fprintf(stderr,"\t>> \"neighbour: addr %u.%u.%u.%u:%u\"\n",
              ((unsigned char *)&faddr)[0],
              ((unsigned char *)&faddr)[1],
              ((unsigned char *)&faddr)[2],
              ((unsigned char *)&faddr)[3],
              fport));
#endif

    /*  Ignore if same as myself.  can occur in very small rings. */
    /*  Equivalence between this host's IP addr and "localhost".  */
    if (fport == PORT &&
        (faddr == my_ipaddr.s_addr || faddr == my_ipaddr_alt.s_addr))
      goto skip;

    /*  If already in local list: ignore but move insert point forward  */

    found = (peer_t *)is_neighbour(faddr,fport);

    if (found) {
      insert_point = found;
      goto skip;
    }

    /*
     *  Apparently a new neighbour, so we are interested in it.
     *  Ask for its pubkey...
     */

    _write_(reqst->csock,"pubkey",strlen("pubkey")+1,rc,break);

#ifdef DEBUG_R
    S(fprintf(stderr, "\t<< pubkey\n"));
#endif

    _read_(reqst->csock,reqst->buf,KEYBUFLEN,rc,TIMEOUT,break);
    sscanf(reqst->buf, "%50000s %100s",keymod,keyexp);

    /*  Build a descriptor for the new neighbour  */

    _do_(_calloc_(f,sizeof(peer_t)),!f,"neblo_neighlist(): f calloc");
    _pthmutexinit_(&f->lock);

    f->addr.s_addr = faddr;
    f->port = fport;

    _do_(_RSA_new_(f->pubkey),!f->pubkey,
         "neblo_neighlist(): RSA_new of f->pubkey");
    _convert_pubkey_(f->pubkey,keymod,keyexp,/**/);

    if (!(f->pubkey->n && f->pubkey->e)) {
      /*  Bad pubkey for this neighbour; discard and go on.  */
      _RSA_free_(f->pubkey);
      _free_(f);
      goto skip;
    }

    f->status = TOCALL;

    /* All other fields are zero/NULL by calloc */

    /* Insert new neighbour in current insert point, update the insert point */

    pthread_mutex_lock(&ring->lock);
    f->next = insert_point->next;
    insert_point->next = f;
    insert_point = f;
    if (f->next == NULL)
      /*  Insertion occurred at tail, so update tail as well  */
      ring->neigh_list_tail = f;
    pthread_mutex_unlock(&ring->lock);

skip:

    _write_(reqst->csock,"ok",strlen("ok")+1,rc,break);

#ifdef DEBUG_R
    S(fprintf(stderr, "\t<< ok\n"));
#endif

  }

exit:

  _free_(keymod); _free_(keyexp);
  return rc;

}
